@@ -59,6 +59,17 @@ dwell_restore_single_click (MTClosure *mt)
     mt_service_set_clicktype (mt->service, DWELL_CLICK_TYPE_SINGLE, NULL);
 }
 
+static gboolean
+dwell_b1r_timeout (gpointer data)
+{
+    gint x, y;
+
+    gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);
+    SPI_generateMouseEvent (x, y, "b1r");
+
+    return FALSE;
+}
+
 static void
 dwell_do_pointer_click (MTClosure *mt, gint x, gint y)
 {
@@ -68,16 +79,28 @@ dwell_do_pointer_click (MTClosure *mt, gint x, gint y)
 
     switch (clicktype) {
     case DWELL_CLICK_TYPE_SINGLE:
-	SPI_generateMouseEvent (x, y, "b1c");
+	SPI_generateMouseEvent (x, y, "b1p");
+	/* Wait a few msecs before releasing the button again.
+	 * This allows GOK to see our clicks.
+	 */
+	g_timeout_add (60, dwell_b1r_timeout, NULL);
 	break;
     case DWELL_CLICK_TYPE_DOUBLE:
 	SPI_generateMouseEvent (x, y, "b1d");
 	dwell_restore_single_click (mt);
 	break;
     case DWELL_CLICK_TYPE_DRAG:
-	SPI_generateMouseEvent (x, y, "b1p");
-	mt->dwell_drag_started = TRUE;
-	mt_cursor_set (GDK_FLEUR);
+	if (!mt->dwell_drag_started) {
+	    SPI_generateMouseEvent (x, y, "b1p");
+	    mt->dwell_drag_started = TRUE;
+	    mt_cursor_set (GDK_FLEUR);
+	}
+	else {
+	    SPI_generateMouseEvent (x, y, "b1r");
+	    mt->dwell_drag_started = FALSE;
+	    mt_cursor_set (GDK_LEFT_PTR);
+	    dwell_restore_single_click (mt);
+	}
 	break;
     case DWELL_CLICK_TYPE_RIGHT:
 	SPI_generateMouseEvent (x, y, "b3c");
@@ -88,13 +111,14 @@ dwell_do_pointer_click (MTClosure *mt, gint x, gint y)
 }
 
 static inline gboolean
-within_tolerance (MTClosure *mt, gint x, gint y)
+below_threshold (MTClosure *mt, gint x, gint y)
 {
-    gdouble distance;
+    gint dx, dy;
 
-    distance = sqrt(pow(x - mt->pointer_x, 2.) + pow(y - mt->pointer_y, 2.));
+    dx = x - mt->pointer_x;
+    dy = y - mt->pointer_y;
 
-    return distance < mt->threshold;
+    return (dx * dx + dy * dy) < (mt->threshold * mt->threshold);
 }
 
 static gboolean
@@ -102,10 +126,10 @@ analyze_direction (MTClosure *mt, gint x, gint y)
 {
     gint gd, i, dx, dy;
 
-    dx = ABS(x - mt->pointer_x);
-    dy = ABS(y - mt->pointer_y);
+    dx = ABS (x - mt->pointer_x);
+    dy = ABS (y - mt->pointer_y);
 
-    if (within_tolerance (mt, x, y))
+    if (below_threshold (mt, x, y))
 	return FALSE;
 
     /* find direction */
@@ -208,45 +232,29 @@ dwell_stop_gesture (MTClosure *mt)
 static void
 dwell_timer_finished (MtTimer *timer, gpointer data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
     gint x, y;
 
     gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);
     mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 
-    /* stop active drag */
-    if (mt->dwell_drag_started) {
-	SPI_generateMouseEvent (x, y, "b1r");
-
-	mt->dwell_drag_started = FALSE;
-	mt_cursor_set (GDK_LEFT_PTR);
-	dwell_restore_single_click (mt);
-
-	return;
-    }
-
-    switch (mt->dwell_mode) {
-    case DWELL_MODE_CTW:
+    if (mt->dwell_mode == DWELL_MODE_CTW)
 	dwell_do_pointer_click (mt, x, y);
-	break;
-    case DWELL_MODE_GESTURE:
+    else if (mt->dwell_mode == DWELL_MODE_GESTURE) {
 	if (mt->dwell_gesture_started) {
 	    dwell_stop_gesture (mt);
-
 	    if (analyze_direction (mt, x, y))
 		dwell_do_pointer_click (mt, mt->pointer_x, mt->pointer_y);
 	}
 	else
 	    dwell_start_gesture (mt);
-    default:
-	break;
     }
 }
 
 static gboolean
 right_click_timeout (gpointer data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     SPI_generateMouseEvent (mt->pointer_x, mt->pointer_y, "b3c");
 
@@ -256,7 +264,7 @@ right_click_timeout (gpointer data)
 static void
 delay_timer_finished (MtTimer *timer, gpointer data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 
@@ -270,16 +278,15 @@ delay_timer_finished (MtTimer *timer, gpointer data)
 static void
 spi_motion_event (const AccessibleEvent *event, void *data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (mt->dwell_enabled) {
-	if (!within_tolerance (mt, event->detail1, event->detail2) &&
+	if (!below_threshold (mt, event->detail1, event->detail2) &&
 	    !mt->dwell_gesture_started) {
 	    mt->pointer_x = (gint) event->detail1;
 	    mt->pointer_y = (gint) event->detail2;
 	    mt_timer_start (mt->dwell_timer);
 	}
-
 	if (mt->dwell_gesture_started) {
 	    if (mt->x_old > -1 && mt->y_old > -1)
 		draw_line (mt->pointer_x, mt->pointer_y, mt->x_old, mt->y_old);
@@ -287,13 +294,12 @@ spi_motion_event (const AccessibleEvent *event, void *data)
 	    draw_line (mt->pointer_x, mt->pointer_y,
 		       event->detail1, event->detail2);
 
-	    mt->x_old = event->detail1;
-	    mt->y_old = event->detail2;
+	    mt->x_old = (gint) event->detail1;
+	    mt->y_old = (gint) event->detail2;
 	}
     }
-
     if (mt_timer_is_running (mt->delay_timer)) {
-	if (!within_tolerance (mt, event->detail1, event->detail2)) {
+	if (!below_threshold (mt, event->detail1, event->detail2)) {
 	    mt_timer_stop (mt->delay_timer);
 	    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 	}
@@ -303,15 +309,14 @@ spi_motion_event (const AccessibleEvent *event, void *data)
 static void
 spi_button_event (const AccessibleEvent *event, void *data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (g_str_equal (event->type, "mouse:button:1p")) {
 	if (mt->delay_enabled) {
-	    gdk_display_get_pointer (gdk_display_get_default (), NULL,
-				     &mt->pointer_x, &mt->pointer_y, NULL);
+	    mt->pointer_x = (gint) event->detail1;
+	    mt->pointer_y = (gint) event->detail2;
 	    mt_timer_start (mt->delay_timer);
 	}
-
 	if (mt->dwell_gesture_started)
 	    dwell_stop_gesture (mt);
     }
@@ -358,7 +363,7 @@ cursor_overlay_time (guchar  *image,
     return TRUE;
 }
 
-void
+static void
 mt_update_cursor (MtCursor *cursor, MtTimer *timer, gdouble time)
 {
     guchar *image;
@@ -390,7 +395,7 @@ mt_update_cursor (MtCursor *cursor, MtTimer *timer, gdouble time)
 static void
 delay_timer_tick (MtTimer *timer, gdouble time, gpointer data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (mt->animate_cursor && mt->cursor != NULL)
 	mt_update_cursor (mt->cursor, timer, time);
@@ -399,7 +404,7 @@ delay_timer_tick (MtTimer *timer, gdouble time, gpointer data)
 static void
 dwell_timer_tick (MtTimer *timer, gdouble time, gpointer data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (mt->animate_cursor && mt->cursor != NULL)
 	mt_update_cursor (mt->cursor, timer, time);
@@ -409,7 +414,7 @@ static void
 cursor_cache_cleared (MtCursorManager *manager,
 		      gpointer         data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     mt->cursor = mt_cursor_manager_current_cursor (manager);
 }
@@ -419,7 +424,7 @@ cursor_changed (MtCursorManager *manager,
 		const gchar     *name,
 		gpointer         data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (!mt->dwell_gesture_started)
 	mt->override_cursor = !g_str_equal (name, "left_ptr");
@@ -439,7 +444,7 @@ gconf_value_changed (GConfClient *client,
 		     GConfValue  *value,
 		     gpointer     data)
 {
-    MTClosure *mt = (MTClosure *) data;
+    MTClosure *mt = data;
 
     if (g_str_equal (key, OPT_THRESHOLD) && value->type == GCONF_VALUE_INT)
 	mt->threshold = gconf_value_get_int (value);
@@ -543,8 +548,10 @@ accessibility_enabled (MTClosure *mt,
 	}
 	else {
 	    /* reset the selected option again */
-	    gconf_client_set_bool (mt->client, OPT_DELAY, FALSE, NULL);
-	    gconf_client_set_bool (mt->client, OPT_DWELL, FALSE, NULL);
+	    if (gconf_client_get_bool (mt->client, OPT_DELAY, NULL))
+		gconf_client_set_bool (mt->client, OPT_DELAY, FALSE, NULL);
+	    if (gconf_client_get_bool (mt->client, OPT_DWELL, NULL))
+		gconf_client_set_bool (mt->client, OPT_DWELL, FALSE, NULL);
 	}
 	return FALSE;
     }
@@ -620,7 +627,7 @@ main (int argc, char **argv)
 	{"enable-secondary", 0, 0, G_OPTION_ARG_NONE, &delay_click,
 	    _("Enable simulated secondary click"), 0},
 	{"dwell-time", 0, 0, G_OPTION_ARG_DOUBLE, &dwell_time,
-	    _("Time to wait before a dwell click"), "[0.5-3.0]"},
+	    _("Time to wait before a dwell click"), "[0.2-3.0]"},
 	{"secondary-time", 0, 0, G_OPTION_ARG_DOUBLE, &delay_time,
 	    _("Time to wait before a simulated secondary click"), "[0.5-3.0]"},
 	{"dwell-mode", 'm', 0, G_OPTION_ARG_STRING, &mode,
@@ -680,6 +687,7 @@ main (int argc, char **argv)
 	MtCursorManager *manager;
 	AccessibleEventListener *bl, *ml;
 	gint spi_status;
+	gint spi_leaks;
 
 	if (mt_pidfile_create () < 0)
 	    return 1;
@@ -697,6 +705,7 @@ main (int argc, char **argv)
 	if (!mt)
 	    goto FINISH;
 
+	spi_leaks = 0;
 	spi_status = SPI_init ();
 	if (!accessibility_enabled (mt, spi_status)) {
 	    mt_closure_free (mt);
@@ -717,7 +726,7 @@ main (int argc, char **argv)
 	    gconf_client_set_bool (mt->client, OPT_DELAY, TRUE, NULL);
 	if (delay_time >= .5 && delay_time <= 3.)
 	    gconf_client_set_float (mt->client, OPT_DELAY_T, delay_time, NULL);
-	if (dwell_time >= .5 && dwell_time <= 3.)
+	if (dwell_time >= .2 && dwell_time <= 3.)
 	    gconf_client_set_float (mt->client, OPT_DWELL_T, dwell_time, NULL);
 	if (threshold >= 0 && threshold <= 30)
 	    gconf_client_set_int (mt->client, OPT_THRESHOLD, threshold, NULL);
@@ -755,14 +764,17 @@ main (int argc, char **argv)
 
 	CLEANUP:
 	    SPI_deregisterGlobalEventListenerAll (bl);
-	    AccessibleDeviceListener_unref (bl);
+	    AccessibleEventListener_unref (bl);
 	    SPI_deregisterGlobalEventListenerAll (ml);
 	    AccessibleEventListener_unref (ml);
-	    SPI_exit ();
+	    spi_leaks = SPI_exit ();
 	    mt_closure_free (mt);
 	FINISH:
 	    mt_pidfile_remove ();
 	    g_object_unref (program);
+
+	if (spi_leaks)
+	    g_warning ("AT-SPI reported %i leaks", spi_leaks);
     }
     return 0;
 }
